import logzero
import pybind11

from clang.cindex import TranslationUnit as TU

from .utils import get_index, get_includes


def parse_tu(
    path,
    input_folder,
    prefix=None,
    platform_includes=[],
    args=[
        "-x",
        "c++",
        "-std=c++17",
        "-D__CODE_GENERATOR__",
        "-Wno-deprecated-declarations",
    ],
    parsing_header="",
    tu_parsing_header="",
    platform_parsing_header="",
    target_platform=None,
):
    """Run a translation unit thorugh clang
    """

    args.append(f"-I{pybind11.get_include()}")
    args.append(f"-I{input_folder}")

    if target_platform == "Windows":
        args.append("--target=x86_64-pc-windows-msvc")
        args.append("-fms-compatibility")
        args.append("-fms-extensions")

    if prefix:
        args.append(f"--sysroot={prefix}")

    for inc in get_includes():
        args.append(f"-I{inc}")

    for inc in platform_includes:
        args.append(f"-I{inc}")

    ix = get_index()

    with open(path) as f:
        src = f.read()

    # skip possible invisible BOM character which would lead to clang error later
    if src[0] == "\ufeff":
        src = src[1:]

    dummy_code = (
        f"{parsing_header}\n{platform_parsing_header}\n{
            tu_parsing_header}\n{src}"
    )
    tr_unit = ix.parse(
        "dummy.cxx",
        args,
        unsaved_files=[("dummy.cxx", dummy_code)],
        options=TU.PARSE_INCOMPLETE,
    )

    diag = list(tr_unit.diagnostics)
    if diag:
        logzero.logger.warning(path)
        for d in diag:
            logzero.logger.warning(d)

    tr_unit.path = ("dummy.cxx", path.name)

    return tr_unit



from schema import Schema, Optional

method_schema = Schema(
    {
        "body": str,
        Optional("help", default=None): str,
        Optional("arguments", default=[]): [str],
        Optional("return_policy", default=None): str,
    }
)

function_schema = method_schema

class_schema = Schema(
    {
        Optional("exclude_constructors", default=[]): [int],
        Optional("additional_constructors", default=[]): [method_schema],
        Optional("additional_methods", default={}): {str: method_schema},
        Optional("additional_static_methods", default={}): {str: method_schema},
    }
)

template_schema = Schema({Optional("exclude_constructors", default=[]): [int]})

module_schema = Schema(
    {
        Optional("module_parsing_header", default=""): str,
        Optional("parsing_headers", default={}): {str: str},
        Optional("include_header_pre_top", default=None): str,
        Optional("include_header_pre", default=None): str,
        Optional("include_body_pre", default=None): str,
        Optional("exclude_functions", default=[]): [str],
        Optional("exclude_classes", default=[]): [str],
        Optional("exclude_class_templates", default=[]): [str],
        Optional("exclude_methods", default=[]): [str],
        Optional("exclude_class_template_methods", default=[]): [str],
        Optional("exclude_typedefs", default=[]): [str],
        Optional("include_body_post", default=None): str,
        Optional("include_body_template_post", default=None): str,
        Optional("include_header_post", default=None): str,
        Optional("template_specializations", default=[]): [str],
        Optional("Classes", default={}): {str: class_schema},
        Optional("Templates", default={}): {str: template_schema},
        Optional("additional_functions", default={}): {str: function_schema},
    }
)

platform_settings = Schema(
    {
        "symbols": str,
        Optional("modules", default=[]): [str],
        Optional("includes", default=[]): [str],
        Optional("prefix", default=None): str,
        Optional("parsing_header", default=""): str,
        Optional("exclude_classes", default=[]): [str],
    }
)

global_schema = Schema(
    {
        "name": str,
        "input_folder": str,
        "output_folder": str,
        "pats": [str],
        "modules": [str],
        Optional("exclude", default=[]): [str],
        Optional("exceptions", default=[]): [str],
        Optional("additional_files", default=[]): [str],
        Optional("template_path", default=None): str,
        Optional("exclude_namespaces", default=[]): [str],
        "module_mapping": str,
        "Operators": {str: [str]},
        "Extras": {
            Optional("include_pre", default=None): str,
            Optional("include_post", default=None): str,
        },
        Optional("byref_types", default=[]): [str],
        Optional("byref_types_smart_ptr", default=[]): [str],
        Optional("parsing_header", default=""): str,
        Optional("Linux", default=None): platform_settings,
        Optional("Windows", default=None): platform_settings,
        Optional("OSX", default=None): platform_settings,
        Optional("FreeBSD", default=None): platform_settings,
        Optional("Attributes", default={}): {str: str},
        Optional("Modules", default=None): {str: module_schema},
    }
)



from typing import List, Tuple, Any, Mapping, Optional
from itertools import chain
from dataclasses import dataclass, field, astuple

from clang.cindex import (
    CursorKind,
    TypeKind,
    AccessSpecifier,
    Cursor,
    Type,
    PrintingPolicy,
)
from path import Path

from .type_parser import parse_type
from .translation_unit import parse_tu
from .utils import current_platform

EXCLUDE_NS: List[str] = []


def paths_approximately_equal(p1: str, p2: str):
    """Approximate path equality. This is due to
    """
    return any([Path(p1).name.split(".")[0] == Path(p).name.split(".")[0] for p in p2])


def is_public(el):

    return (el.acess_specifier == AccessSpecifier.PUBLIC) or (
        el.access_specifier is None
    )


def get_symbols(
    tu,
    kind,
    ignore_forwards=True,
    search_in=(CursorKind.NAMESPACE,),
    exclude_ns: List[str] = [],
):
    """
    Symbols defined locally (i.e. without includes) and are not forward declarations
    Search_in allows to explore nested entities as well.

    """
    tu_path = tu.path

    def _get_symbols(cursor, kind):

        for child in cursor.get_children():
            if (
                paths_approximately_equal(
                    Path(child.location.file.name), tu_path)
                and child.kind == kind
            ):
                if ignore_forwards:
                    if child.get_definition() is None:
                        pass  # forward declaration
                    elif not paths_approximately_equal(
                        Path(child.get_definition().location.file.name), tu_path
                    ):
                        pass  # forward declaration but declared in an include
                    else:
                        yield child  # legitimate
                else:
                    yield child
            if (
                paths_approximately_equal(
                    Path(child.location.file.name), tu_path)
                and child.kind in search_in
                and child.spelling not in exclude_ns
            ):
                for nested in _get_symbols(child, kind):
                    if nested.access_specifier in (
                        AccessSpecifier.PUBLIC,
                        AccessSpecifier.INVALID,
                    ):
                        yield nested

    yield from _get_symbols(tu.cursor, kind)


def get_forward_declarations(tu):
    """Get all symbols that are forward declared"""
    tu_path = tu.path

    for child in tu.cursor.get_children():
        if paths_approximately_equal(Path(child.location.file.name), tu_path):
            if child.get_definition() is None:
                yield child


def get_all_symbols(tu, kind):
    """All defined symbols of given kind
    """
    tu_path = tu.path

    for child in tu.cursor.get_children():
        if (
            paths_approximately_equal(Path(child.location.file.name), tu_path)
            and child.kind is kind
        ):
            yield child


def get_all_symbols_multi(tu, kinds):
    """All defined symbols of given kinds
    """
    tu_path = tu.path

    for child in tu.cursor.get_children():
        if paths_approximately_equal(Path(child.location.file.name), tu_path) and any(
            (child.kind is kind for kind in kinds)
        ):
            yield child


def get_functions(tu):
    """Functions defined locally (i.e. without includes)
    """

    return (
        f
        for f in get_symbols(tu, CursorKind.FUNCTION_DECL, False)
        if "operator" not in f.spelling
    )


def get_function_templates(tu):
    """Function templates defined locally (i.e. without includes)
    """

    return (
        f
        for f in get_symbols(tu, CursorKind.FUNCTION_TEMPLATE, False)
        if "operator" not in f.spelling
    )


def get_operators(tu):
    """Functions defined locally (i.e. without includes)
    """

    return (
        f
        for f in get_symbols(tu, CursorKind.FUNCTION_DECL, False)
        if "operator" in f.spelling
    )


def get_operator_templates(tu):
    """Operator templates defined locally (i.e. without includes)
    """

    return (
        f
        for f in get_symbols(tu, CursorKind.FUNCTION_TEMPLATE, False)
        if "operator" in f.spelling
    )


def get_enums(tu):
    """Enums defined locally (i.e. without includes)
    """

    return get_symbols(tu, CursorKind.ENUM_DECL)


def get_enum_values(cur: Cursor):
    """Gets enum values
    """

    for child in cur.get_children():
        if child.kind is CursorKind.ENUM_CONSTANT_DECL:
            yield child


def get_typedefs(tu):
    """Typedefs defined locally (i.e. without includes)
    """

    return get_symbols(tu, CursorKind.TYPEDEF_DECL)


def get_classes(tu):
    """Classes defined locally (i.e. without includes)
    """

    return chain(
        get_symbols(tu, CursorKind.CLASS_DECL), get_symbols(
            tu, CursorKind.STRUCT_DECL)
    )


def get_class_templates(tu):
    """Class templates defined locally (i.e. without includes)
    """

    return get_symbols(tu, CursorKind.CLASS_TEMPLATE)


def get_namespaces(tu, ignore_ns=[]):
    """Namepspaces defined locally (i.e. without includes).
    """

    for el in get_symbols(tu, CursorKind.NAMESPACE):
        if el.spelling not in ignore_ns+[""]:
            yield el


def get_x(cls, kind):
    """Get children entities of the specified type excluding forward declataions
    """

    for child in cls.get_children():
        if child.kind is kind and child.get_definition():
            yield child


def get_x_multi(cls, kinds):
    """Get children entities of the specified types excluding forward declataions
    """

    for child in cls.get_children():
        if any((child.kind is kind for kind in kinds)) and child.get_definition():
            yield child


def get_xx(cls, kind, access):
    """Get children entities of the specified type with given access specifier
    """

    for child in cls.get_children():
        if child.kind is kind and child.access_specifier is access:
            yield child


def get_template_type_params(cls):
    """Get all template type params
    """

    for t in get_x_multi(
        cls,
        (CursorKind.TEMPLATE_TYPE_PARAMETER,
         CursorKind.TEMPLATE_NON_TYPE_PARAMETER),
    ):
        if len(list(t.get_children())) == 0:
            yield t, None
        else:

            default = "".join([tok.spelling for tok in t.get_tokens()][3:])

            # workaround - needs to be investigated
            if default.endswith(">>"):
                default = default[:-1]

            yield t, default


def get_base_class(c):
    """Get all class-baseclass pairs with public or protected inheritance
    """

    if c.get_definition():
        rv = [
            el
            for el in get_x(c.get_definition(), CursorKind.CXX_BASE_SPECIFIER)
            # how to handle  AccessSpecifier.PROTECTED?
            if el.access_specifier in (AccessSpecifier.PUBLIC,)
        ]
    else:
        rv = [
            el
            for el in get_x(c, CursorKind.CXX_BASE_SPECIFIER)
            # how to handle  AccessSpecifier.PROTECTED?
            if el.access_specifier in (AccessSpecifier.PUBLIC,)
        ]

    if len(rv) == 0:
        return []
    else:
        return [el.type.spelling for el in rv]


def get_inheritance_relations(tu):
    """Inheritance relations pairs
    """

    all_classes = get_x(tu.cursor, CursorKind.CLASS_DECL)

    for c in all_classes:
        yield c.spelling, get_base_class((c))

    all_templates = get_x(tu.cursor, CursorKind.CLASS_TEMPLATE)

    for c in all_templates:
        yield c.spelling, get_base_class((c))


def get_public_fields(cls):
    """Public methods of a given class
    """

    for child in get_xx(cls, CursorKind.FIELD_DECL, AccessSpecifier.PUBLIC):
        yield child


def get_public_enums(cls):
    """Public enums of a given class
    """

    for child in get_xx(cls, CursorKind.ENUM_DECL, AccessSpecifier.PUBLIC):
        yield child


def get_public_methods(cls):
    """Public methods of a given class
    """

    for child in get_xx(cls, CursorKind.CXX_METHOD, AccessSpecifier.PUBLIC):
        if not child.is_static_method() and not child.spelling.startswith("operator"):
            yield child


def get_protected_pure_virtual_methods(cls):
    """Protected pure virtual methods of a given class
    """

    for child in get_xx(cls, CursorKind.CXX_METHOD, AccessSpecifier.PROTECTED):
        if (
            not child.is_static_method()
            and not child.spelling.startswith("operator")
            and child.is_pure_virtual_method()
        ):
            yield child


def get_private_pure_virtual_methods(cls):
    """Private pure virtual methods of a given class
    """

    for child in get_xx(cls, CursorKind.CXX_METHOD, AccessSpecifier.PRIVATE):
        if (
            not child.is_static_method()
            and not child.spelling.startswith("operator")
            and child.is_pure_virtual_method()
        ):
            yield child


def get_public_static_methods(cls):
    """Public static methods of a given class
    """

    for child in get_xx(cls, CursorKind.CXX_METHOD, AccessSpecifier.PUBLIC):
        if child.is_static_method() and not child.spelling.startswith("operator"):
            yield child


def get_public_operators(cls):
    """Public operators of a given class
    """

    for child in get_xx(cls, CursorKind.CXX_METHOD, AccessSpecifier.PUBLIC):
        if not child.is_static_method() and child.spelling.startswith("operator"):
            yield child


def get_public_static_operators(cls):
    """Public static operators of a given class
    """

    for child in get_xx(cls, CursorKind.CXX_METHOD, AccessSpecifier.PUBLIC):
        if child.is_static_method() and child.spelling.startswith("operator"):
            yield child


def get_public_constructors(cls):
    """Public constructors of a given class
    """

    for child in get_xx(cls, CursorKind.CONSTRUCTOR, AccessSpecifier.PUBLIC):
        yield child


def get_private_constructors(cls):
    """Private constructors of a given class
    """

    for child in get_xx(cls, CursorKind.CONSTRUCTOR, AccessSpecifier.PRIVATE):
        yield child


def get_protected_constructors(cls):
    """Protected constructors of a given class
    """

    for child in get_xx(cls, CursorKind.CONSTRUCTOR, AccessSpecifier.PROTECTED):
        yield child


def get_public_destructors(cls):
    """Public destructors of a given class
    """

    for child in get_xx(cls, CursorKind.DESTRUCTOR, AccessSpecifier.PUBLIC):
        yield child


def get_private_destructors(cls):
    """Private destructors of a given class
    """

    for child in get_xx(cls, CursorKind.DESTRUCTOR, AccessSpecifier.PRIVATE):
        yield child


def get_protected_destructors(cls):
    """Private destructors of a given class
    """

    for child in get_xx(cls, CursorKind.DESTRUCTOR, AccessSpecifier.PROTECTED):
        yield child


def get_free_method_definitions(tu):
    """Free method definitions
    """

    return (
        el
        for el in get_symbols(tu, CursorKind.CXX_METHOD)
        if not el.is_static_method() and el.access_specifier == AccessSpecifier.PUBLIC
    )


def get_template_arg_dict(cur) -> dict[str, str]:
    """
    Retrun a dict mapping canonical to real template names.
    """

    rv = {}

    for ch in cur.get_children():
        if ch.kind == CursorKind.TEMPLATE_TYPE_PARAMETER:
            rv[ch.type.get_canonical().spelling] = ch.spelling
        else:
            break

    return rv


def get_named_type(T: Type) -> Type:
    """
    Strips const and &/ptr from Type.
    """

    rv = T

    if rv.kind in (TypeKind.LVALUEREFERENCE, TypeKind.POINTER):
        rv = rv.get_pointee()

    if rv.kind == TypeKind.ELABORATED:
        rv = rv.get_named_type()

    return rv


class BaseInfo(object):
    """Base class for the info objects
    """

    name: str
    comment: str

    def __init__(self, cur: Cursor):

        self.name = cur.spelling
        self.comment = cur.brief_comment


class FieldInfo(BaseInfo):
    """Container for field parsing reults
    """

    type: str
    const: bool
    pod: bool

    def __init__(self, cur: Cursor):

        super(FieldInfo, self).__init__(cur)

        self.type = cur.type.spelling
        self.const = cur.type.is_const_qualified()
        self.pod = cur.type.is_pod()


class EnumInfo(BaseInfo):
    """Container for enum parsing results
    """

    comment: str
    values: List[str]
    anonymous: bool

    def __init__(self, cur: Cursor):

        super(EnumInfo, self).__init__(cur)

        self.comment = cur.brief_comment
        self.values = [el.spelling for el in get_enum_values(cur)]
        self.anonymous = False
        self.name = cur.type.spelling

        if any(x in self.name for x in ["anonymous", "unnamed"]):
            self.anonymous = True
            self.name = "::".join(self.name.split(
                "::")[:-1])  # get rid of anonymous


@dataclass
class ArgInfo:

    arg_name: str
    arg_type: str
    arg_defualt: str
    arg_py_name: str = field(init=False)

    def __post_init__(self):

        if self.arg_name in ('def',):  # extend with more reserved words
            self.arg_py_name = f'{self.arg_name}_'
        else:
            self.arg_py_name = self.arg_name

    def __iter__(self):

        return iter(astuple(self))


class FunctionInfo(BaseInfo):
    """Container for function parsing results
    """

    namespace: Optional[str]
    full_name: str
    mangled_name: str
    return_type: str
    inline: bool
    pointer_by_ref: bool
    args: List[ArgInfo]
    default_value_types: List[str]

    KIND_DICT = {
        TypeKind.LVALUEREFERENCE: " &",
        TypeKind.RVALUEREFERENCE: " &&",
        TypeKind.POINTER: " *",
    }

    def __init__(self, cur: Cursor):

        super(FunctionInfo, self).__init__(cur)

        self.comment = cur.brief_comment
        self.full_name = cur.displayname
        self.mangled_name = cur.mangled_name
        self.return_type = self._underlying_type(cur.result_type, cur)

        if cur.semantic_parent.kind == CursorKind.NAMESPACE:
            self.namespace = cur.semantic_parent.spelling
        else:
            self.namespace = None

        self.inline = (
            cur.get_definition().is_inline() if cur.get_definition() else False
        )
        self.pointer_by_ref = any(
            self._pointer_by_ref(el) for el in cur.get_arguments()
        )
        self.args = [
            ArgInfo(el.spelling, self._underlying_type(
                el, cur), self._default_value(el))
            for el in cur.get_arguments()
        ]
        self.default_value_types = [
            self._underlying_type(el, cur, False)
            for el in cur.get_arguments()
            if self._default_value(el)
        ]

    def _pointer_by_ref(self, cur: Cursor) -> bool:
        """Check is type is a Pointer passed by reference
        """

        rv = False
        t = cur.type

        # if passed by ref
        if t.kind == TypeKind.LVALUEREFERENCE:
            tp = t.get_pointee().get_canonical()
            # if underlying type is a pointer
            if tp.kind in (TypeKind.POINTER,):
                rv = True

        return rv

    def _underlying_type(
        self, cur: Cursor | Type, ctx: Cursor, add_qualifiers=True
    ) -> str:
        """Tries to resolve the underlying type. Needed for typedefed templates.
        """

        T = cur.type if isinstance(cur, Cursor) else cur

        ptr = self.KIND_DICT.get(T.kind, "")

        policy = PrintingPolicy.create(ctx)

        # if lvaule,rvalue or pointer type
        if ptr:
            const_ptr = " const " if T.is_const_qualified() else ""
            pointee = T.get_pointee()
            const = " const " if pointee.is_const_qualified() else ""
            typ = pointee
        else:
            const_ptr = ""
            const = " const " if T.is_const_qualified() else ""
            typ = T

        if typ.kind == TypeKind.ELABORATED:
            typ = typ.get_named_type()

        rv = typ.pretty_printed(policy)

        # strip constness if still present
        if typ.is_const_qualified():
            rv = rv.removeprefix("const ")

        # handle typdefs that are not pods
        if typ.kind == TypeKind.TYPEDEF and not typ.is_pod():
            decl = typ.get_declaration()
            # replace the spelling with displayname to add template arguments - this is very brittle
            rv = rv.replace(
                decl.semantic_parent.spelling + "::",
                decl.semantic_parent.displayname + "::",
            )

            if decl.semantic_parent.kind != CursorKind.TRANSLATION_UNIT:
                rv = "typename " + rv

        # additional postprocessing related to templates
        if typ.kind == TypeKind.UNEXPOSED:
            typ_can = typ.get_canonical().spelling

            # handle missing std::
            if typ_can.startswith("std::") and not rv.startswith("std::"):
                rv = "std::" + rv

            # expand template argument names if needed
            n_args = typ.get_num_template_arguments()
            if n_args != -1:
                # construct a short:long name dict
                args = {}
                for ix in range(n_args):
                    arg = typ.get_template_argument_type(ix)
                    args[arg.spelling] = self._underlying_type(arg, ctx)

                # transform the result
                for k, v in args.items():
                    rv = rv.replace(k, v)

        if add_qualifiers:
            rv = const + rv + ptr + const_ptr

        # strip possible opencascade::handle<T> when
        if not add_qualifiers:
            rv = parse_type(rv)

        return rv

    def _default_value(self, cur: Cursor) -> Optional[str]:
        """Tries to extract default value
        """

        rv = None
        tokens = [t.spelling for t in cur.get_tokens()]
        if "=" in tokens:
            rv = " ".join(tokens[tokens.index("=") + 1:])

            # handle default initalization of complex types
            if "{ }" == rv:
                rv = f"{get_named_type(cur.type).spelling}{rv}"

        return rv


class MethodInfo(FunctionInfo):
    """Container for method parsing results
    """

    const: bool
    virtual: bool
    pure_virtual: bool

    def __init__(self, cur: Cursor):

        super(MethodInfo, self).__init__(cur)

        self.const = cur.is_const_method()
        self.virtual = cur.is_virtual_method()
        self.pure_virtual = cur.is_pure_virtual_method()


class ConstructorInfo(MethodInfo):
    """Container for constructor parsing results
    """

    pass


class DestructorInfo(FunctionInfo):
    """Container for destructor parsing results
    """

    pass


class ClassInfo(object):
    """Container for class parsing results
    """

    name: str
    comment: str
    abstract: bool

    constructors: List[ConstructorInfo]
    nonpublic_constructors: List[ConstructorInfo]

    fields: List[FieldInfo]

    enums: List[EnumInfo]

    methods: List[MethodInfo]
    protected_virtual_methods: List[MethodInfo]
    private_virtual_methods: List[MethodInfo]
    static_methods: List[MethodInfo]
    static_methods_byref: List[MethodInfo]
    methods_byref: List[MethodInfo]
    methods_return_byref: List[MethodInfo]

    operators: List[MethodInfo]
    static_operators: List[MethodInfo]

    destructors: List[DestructorInfo]
    nonpublic_destructors: List[DestructorInfo]

    ptr: Any  # is this needed?

    superclass: List[str]
    rootclass: List[str]
    superclasses: List[str]

    methods_dict: Mapping[str, MethodInfo]
    protected_virtual_methods_dict: Mapping[str, MethodInfo]
    private_virtual_methods_dict: Mapping[str, MethodInfo]

    def __init__(self, cur: Cursor):

        self.name = cur.type.spelling
        self.comment = cur.brief_comment
        self.abstract = cur.is_abstract_record()

        self.constructors = self.filter_rvalues(
            (ConstructorInfo(el) for el in get_public_constructors(cur))
        )
        self.nonpublic_constructors = [
            ConstructorInfo(el) for el in get_private_constructors(cur)
        ] + [ConstructorInfo(el) for el in get_protected_constructors(cur)]

        self.fields = [FieldInfo(el) for el in get_public_fields(cur)]
        self.enums = [EnumInfo(el) for el in get_public_enums(cur)]

        self.methods = self.filter_rvalues(
            (MethodInfo(el) for el in get_public_methods(cur))
        )
        self.protected_virtual_methods = self.filter_rvalues(
            (MethodInfo(el) for el in get_protected_pure_virtual_methods(cur))
        )
        self.private_virtual_methods = self.filter_rvalues(
            (MethodInfo(el) for el in get_private_pure_virtual_methods(cur))
        )

        self.static_methods = self.filter_rvalues(
            (MethodInfo(el) for el in get_public_static_methods(cur))
        )

        self.static_methods_byref = []
        self.methods_byref = []
        self.methods_return_byref = []

        self.operators = [MethodInfo(el) for el in get_public_operators(cur)]
        self.static_operators = [
            MethodInfo(el) for el in get_public_static_operators(cur)
        ]

        self.destructors = [DestructorInfo(el)
                            for el in get_public_destructors(cur)]
        self.nonpublic_destructors = [
            DestructorInfo(el) for el in get_private_destructors(cur)
        ] + [DestructorInfo(el) for el in get_protected_destructors(cur)]

        self.ptr = None
        self.superclass = []
        self.rootclass = []
        self.superclasses = []

        self.methods_dict = {m.name: m for m in self.methods}
        self.protected_virtual_methods_dict = {
            m.name: m for m in self.protected_virtual_methods
        }
        self.private_virtual_methods_dict = {
            m.name: m for m in self.private_virtual_methods
        }

    def filter_rvalues(self, funcs):

        return [f for f in funcs if not any(("&&" in arg for _, arg, _, _ in f.args))]

    def extend_defintion(self, other):

        new_comment = ""
        if self.comment:
            new_comment += self.comment
        if other.comment:
            new_comment += other.comment
        self.comment = new_comment

        self.constructors += other.constructors
        self.nonpublic_constructors += other.nonpublic_constructors
        self.methods += other.methods
        self.protected_virtual_methods += other.protected_virtual_methods
        self.private_virtual_methods += other.private_virtual_methods
        self.static_methods += other.static_methods
        self.operators += other.operators
        self.static_operators += other.static_operators
        self.destructors += other.destructors
        self.nonpublic_destructors += other.nonpublic_destructors

        self.fields += other.fields
        self.enums += other.enums

        self.methods_dict = {**self.methods_dict, **other.methods_dict}
        self.protected_virtual_methods_dict = {
            **self.protected_virtual_methods_dict,
            **other.protected_virtual_methods_dict,
        }
        self.private_virtual_methods_dict = {
            **self.protected_virtual_methods_dict,
            **other.protected_virtual_methods_dict,
        }


class ClassTemplateInfo(ClassInfo):

    type_params: List[Tuple[Optional[str], str, str]]

    def __init__(self, cur: Cursor):
        super(ClassTemplateInfo, self).__init__(cur)
        self.name = cur.spelling
        self.type_params = [
            (
                None if el.spelling == el.type.spelling else el.type.spelling,
                el.spelling,
                default,
            )
            for el, default in get_template_type_params(cur)
        ]


class TypedefInfo(BaseInfo):

    type: str
    pod: bool
    template_base: List[str]
    template_args: List[str]

    def __init__(self, cur: Cursor):

        super(TypedefInfo, self).__init__(cur)

        t = cur.underlying_typedef_type

        self.type = t.spelling
        self.pod = t.is_pod()

        if not self.pod:
            self.template_base = [
                ch.spelling
                for ch in cur.get_children()
                if ch.kind == CursorKind.TEMPLATE_REF
            ]
            self.template_args = [
                ch.spelling
                for ch in cur.get_children()
                if ch.kind == CursorKind.TYPE_REF
            ]


class ForwardInfo(BaseInfo):

    pass


class HeaderInfo(object):
    """Container for header parsing results
    """

    name: str
    short_name: str
    dependencies: List[str]
    classes: Mapping[str, ClassInfo]
    class_templates: Mapping[str, ClassTemplateInfo]
    functions: List[FunctionInfo]
    operators: List[FunctionInfo]
    enums: List[EnumInfo]
    methods: List[MethodInfo]
    inheritance: Mapping[str, str]
    typedefs: List[TypedefInfo]
    typedef_dict: Mapping[str, str]
    forwards: List[ForwardInfo]
    namespaces: List[str]

    def __init__(self):

        self.name = ""
        self.short_name = ""
        self.dependencies = []
        self.classes = {}
        self.class_templates = {}
        self.funstions = []
        self.operators = []
        self.enums = []
        self.methods = []
        self.inheritance = {}
        self.typedefs = []
        self.typdef_dict = {}
        self.forwards = []
        self.namepspaces = []

    def resolve_inheritance(self, cls):

        inheritance = self.inheritance
        superclasses = cls.superclasses
        rootclass = cls.rootclass

        cls.superclass = inheritance.get(cls.name, [])

        def _resolve_inheritance(node):
            """Recurisvely traverse the inheritance tree"""

            for el in node:
                superclasses.append(el)

                if el in inheritance:
                    _resolve_inheritance(inheritance[el])
                else:
                    rootclass.append(el)

        _resolve_inheritance(cls.superclass)

    def parse(self, path, input_folder, settings, module_name, target_platform):

        module_settings = settings["Modules"].get(module_name)

        if module_settings:
            tu_parsing_header = "\n".join(
                (
                    module_settings["module_parsing_header"],
                    module_settings["parsing_headers"].get(path.name, ""),
                )
            )
        else:
            tu_parsing_header = ""

        tr_unit = parse_tu(
            path,
            input_folder,
            prefix=settings[current_platform()]["prefix"],
            platform_includes=settings[current_platform()]["includes"],
            parsing_header=settings["parsing_header"],
            tu_parsing_header=tu_parsing_header,
            platform_parsing_header=settings[current_platform(
            )]["parsing_header"],
            target_platform=target_platform,
        )

        self.name = path
        self.short_name = path.splitpath()[-1]
        self.dependencies = [
            el.location.file.name for el in tr_unit.get_includes()]
        self.enums = [EnumInfo(el) for el in get_enums(tr_unit)]
        self.functions = [FunctionInfo(el) for el in get_functions(tr_unit)]
        self.operators = [FunctionInfo(el) for el in get_operators(tr_unit)]

        self.classes = {}

        for el in get_classes(tr_unit):
            ci = ClassInfo(el)
            if ci.name in self.classes:
                self.classes[ci.name].extend_defintion(ci)
            else:
                self.classes[ci.name] = ci

        self.class_dict = {k: self.name for k in self.classes}
        self.class_templates = {
            el.displayname: ClassTemplateInfo(el) for el in get_class_templates(tr_unit)
        }
        self.class_template_dict = {k: self.name for k in self.class_templates}
        self.inheritance = {k: v for k,
                            v in get_inheritance_relations(tr_unit) if v}
        self.typedefs = [TypedefInfo(el) for el in get_typedefs(tr_unit)]
        self.typedef_dict = {t.name: self.name for t in self.typedefs}
        self.forwards = [ForwardInfo(el)
                         for el in get_forward_declarations(tr_unit)]

        self.namespaces = [el.spelling for el in get_namespaces(tr_unit)]

        # handle freely defined methods
        methods = [el for el in get_free_method_definitions(tr_unit)]

        # match methods to classes
        for m in methods:
            cls = self.classes.get(m.semantic_parent.displayname, None)
            if cls:
                cls.methods.append(MethodInfo(m))

        # resolve inheritance relations
        for c in self.classes.values():
            self.resolve_inheritance(c)

        for c in self.class_templates.values():
            self.resolve_inheritance(c)

        return tr_unit


def process_header(path, input_folder, settings, module_name, target_platform):
    """Main function from this module
    """

    hi = HeaderInfo()
    hi.parse(path, input_folder, settings, module_name, target_platform)

    return hi


__all__ = [
    process_header,
]

if __name__ == "__main__":

    from os import getenv
    from .utils import init_clang

    init_clang()

    conda_prefix = Path(getenv("CONDA_PREFIX"))

    gp_Ax1 = process_header(conda_prefix / "include" /
                            "opencascade" / "gp_Ax1.hxx")

    for el in gp_Ax1.classes.values():
        print(el.name)

    # try public methods
    for el in el.methods:
        print(el.name, el.args, el.return_type)

    # try enums
    gp_TrsfForm = process_header(
        conda_prefix / "include" / "opencascade" / "gp_TrsfForm.hxx"
    )

    for el in gp_TrsfForm.enums:
        print(el.name)
        print(el.values)

    # try functions
    gp_Vec2d = process_header(
        conda_prefix / "include" / "opencascade" / "gp_Vec2d.hxx")

    for el in gp_Vec2d.functions:
        print(el.name)

    for el in gp_Vec2d.operators:
        print(el.name)

    # try inheritance
    TopoDS = process_header(
        conda_prefix / "include" / "opencascade" / "TopoDS_Solid.hxx"
    )

    for k in TopoDS.inheritance:
        print(k, TopoDS.inheritance[k])



from clang.cindex import Config, Index, Cursor
from ctypes import c_uint
from path import Path
from os import getenv
from sys import platform, prefix

from .cymbal import monkeypatch_cursor

initialized = False
ix = None


def current_platform():

    rv = None
    if platform.startswith("win"):
        rv = "Windows"
    elif platform.startswith("linux"):
        rv = "Linux"
    elif platform.startswith("darwin"):
        rv = "OSX"
    elif platform.startswith("freebsd"):
        rv = "FreeBSD"
    else:
        raise RuntimeError(f"Unsupported platform: {platform}")

    return rv


def on_windows():

    rv = False
    if platform.startswith("win"):
        rv = True

    return rv


def get_includes(rv=[]):

    if rv:
        pass
    elif on_windows():
        rv.append(Path(prefix) / "Library/include/clang/")
    else:
        rv.append(Path(prefix) / "lib/clang/8.0.0/include/")
        rv.append(Path(prefix) / "lib/clang/6.0.1/include/")
        rv.append(Path(prefix) / "lib/clang/9.0.1/include/")
        rv.append(Path(prefix) / "lib/clang/10.0.1/include/")
        rv.append(Path(prefix) / "include/c++/v1/")

    return rv


def init_clang(path=None):

    global initialized, ix

    if not initialized:
        conda_prefix = Path(getenv("CONDA_PREFIX", ""))

        if path:
            pass
        elif platform.startswith("win"):
            path = conda_prefix / "Library" / "bin" / "libclang.dll"
        elif platform.startswith("linux") or platform.startswith("freebsd"):
            path = conda_prefix / "lib" / "libclang.so"
        elif platform.startswith("darwin"):
            path = conda_prefix / "lib" / "libclang.dylib"

        Config.set_library_file(path)

        # Monkeypatch clang
        monkeypatch_cursor("is_virtual", "clang_isVirtualBase", [Cursor], c_uint)

        monkeypatch_cursor(
            "is_inline", "clang_Cursor_isFunctionInlined", [Cursor], c_uint
        )

        monkeypatch_cursor(
            "get_specialization", "clang_getSpecializedCursorTemplate", [Cursor], Cursor
        )

        monkeypatch_cursor(
            "get_template_kind", "clang_getTemplateCursorKind", [Cursor], c_uint
        )

        monkeypatch_cursor(
            "get_num_overloaded_decl", "clang_getNumOverloadedDecls", [Cursor], c_uint
        )

        initialized = True
        ix = Index.create()


def get_index():

    global initialized, ix

    if not initialized:
        init_clang()

    return ix



from pyparsing import Word, Literal, alphas, alphanums, Optional


CONST = Literal("const")
HANDLE = Literal("opencascade::handle<")
TYPE = Word(alphas + "_", alphanums + "_")
CLOSING = Literal(">")
PTR_REF = Literal("&") | Literal("*")

parser = (
    Optional(CONST)
    + Optional(HANDLE)
    + TYPE.setResultsName("type")
    + Optional(CLOSING)
    + Optional(PTR_REF)
)


def parse_type(t):

    return parser.parseString(t).type



import click
import logzero
import pickle

from types import SimpleNamespace
from path import Path

from . import read_settings, parse_modules, transform_modules, render, validate_result
from .utils import get_includes, init_clang


@click.group()
@click.option("-n", "--njobs", default=-2, type=int)
@click.option(
    "-p",
    "--prefix",
    default=".",
    type=click.Path(False, False, True),
    help="Source prefix",
)
@click.option("-v", "--verbose", is_flag=True)
@click.option("-c", "--clean", is_flag=True)
@click.option(
    "-i",
    "--include",
    multiple=True,
    type=click.Path(True, False, True),
    default=[],
    help="additional inlcude paths",
)
@click.option(
    "-l",
    "--libclang",
    type=click.Path(True, True, False),
    default=None,
    help="libclang location",
)
@click.pass_context
def main(ctx, clean, verbose, njobs, include, prefix, libclang):

    if not verbose:
        logzero.logger.setLevel(logzero.logging.INFO)

    if include:
        get_includes.__defaults__ = (include,)

    if libclang:
        init_clang.__defaults__ = (libclang,)

    ctx.obj = SimpleNamespace(
        verbose=verbose, njobs=njobs, clean=clean, prefix=Path(prefix)
    )


@main.command()
@click.argument("configuration")
@click.argument("output")
@click.argument(
    "platform",
    default=None,
    required=False,
    type=click.Choice(("Linux", "Windows", "OSX", "FreeBSD")),
)
@click.pass_obj
def parse(obj, configuration, output, platform=None):

    settings, module_mapping, module_settings = read_settings(configuration)

    with obj.prefix:
        result = parse_modules(
            obj.verbose, obj.njobs, settings, module_mapping, module_settings, platform
        )

    with open(output, "wb") as f:
        pickle.dump(result, f)


@main.command()
@click.argument("configuration")
@click.argument(
    "platform",
    default=None,
    required=False,
    type=click.Choice(("Linux", "Windows", "OSX", "FreeBSD")),
)
@click.argument("input")
@click.argument("output")
@click.pass_obj
def transform(obj, configuration, platform, input, output):

    with open(input, "rb") as f:
        modules = pickle.load(f)

    settings, module_mapping, module_settings = read_settings(configuration)
    modules, class_dict, enum_dict = transform_modules(
        obj.verbose, obj.njobs, settings, module_mapping, module_settings, modules, platform=platform
    )

    with open(output, "wb") as f:
        pickle.dump((modules, class_dict, enum_dict), f)


@main.command()
@click.argument("configuration")
@click.argument(
    "platform",
    default=None,
    required=False,
    type=click.Choice(("Linux", "Windows", "OSX", "FreeBSD")),
)
@click.argument("input")
@click.pass_obj
def generate(obj, configuration, platform, input):

    settings, module_mapping, module_settings = read_settings(configuration)
    out = Path(settings["output_folder"])

    if obj.clean:
        out.rmtree_p()

    with open(input, "rb") as f:
        modules, class_dict, enum_dict = pickle.load(f)

    render(settings, module_settings, modules, class_dict, obj.prefix, platform=platform)

    pre = settings["Extras"]["include_pre"]
    post = settings["Extras"]["include_pre"]

    if pre:
        (obj.prefix / Path(pre)).copy(out)
    if post:
        (obj.prefix / Path(post)).copy(out)


@main.command()
@click.argument("folder")
@click.pass_obj
def validate(obj, folder):

    validate_result(obj.verbose, obj.njobs, folder)


@main.command()
@click.argument("configuration")
@click.argument(
    "platform",
    default=None,
    required=False,
    type=click.Choice(("Linux", "Windows", "OSX", "FreeBSD")),
)
@click.argument("tmp_parsed", default="tmp.pkl")
@click.argument("tmp_filtered", default="tmp_filtered.pkl")
@click.pass_context
def all(ctx, configuration, platform, tmp_parsed, tmp_filtered):

    ctx.invoke(parse, configuration=configuration, output=tmp_parsed, platform=platform)
    ctx.invoke(
        transform, configuration=configuration, input=tmp_parsed, output=tmp_filtered, platform=platform
    )
    ctx.invoke(generate, configuration=configuration, input=tmp_filtered, platform=platform)


if __name__ == "__main__":

    main()



import sys

from jinja2 import Environment, meta, FileSystemLoader
from path import Path


def get_variables(filename):
    env = Environment(loader=FileSystemLoader(Path(filename).dirname()))
    template_source = env.loader.get_source(env, Path(filename).name)[0]
    parsed_content = env.parse(template_source)

    return meta.find_undeclared_variables(parsed_content)


if __name__ == "__main__":

    env = Environment()
    with open(sys.argv[1]) as template:
        env.parse(template.read())



from typing import List, Set, Mapping, Any

from .header import (
    process_header,
    HeaderInfo,
    ClassInfo,
    ClassTemplateInfo,
    TypedefInfo,
    FunctionInfo,
    EnumInfo,
)
from logzero import logger
from toposort import toposort_flatten

from path import Path


class ModuleInfo(object):
    """Container for the whole module
    """

    prefix: str
    name: str
    headers: List[HeaderInfo]

    classes: List[ClassInfo]
    class_dict: Mapping[str, ClassInfo]
    class_templates: List[ClassTemplateInfo]
    class_template_dict: Mapping[str, ClassTemplateInfo]
    typedefs: List[TypedefInfo]
    typedef_dict: Mapping[str, TypedefInfo]
    enums: List[EnumInfo]
    functions: List[FunctionInfo]
    operators: List[FunctionInfo]
    exceptions: List[Any]

    dependencies: Set[Any]
    dependencies_headers: Set[str]

    namespaces: Set[str]

    def get_module_name(self, x):

        return Path(x).splitpath()[-1].split(".")[0].split("_")[0]

    def __init__(self, name, prefix, paths, module_names, settings, target_platform):

        self.prefix = prefix
        self.name = name
        self.headers = []

        logger.debug("Processing headers")

        for p in paths:
            logger.debug(p)
            self.headers.append(process_header(
                p, prefix, settings, name, target_platform))

        self.classes = []
        self.class_dict = {}
        self.class_templates = []
        self.class_template_dict = {}
        self.typedefs = []
        self.typedef_dict = {}
        self.enums = []
        self.functions = []
        self.operators = []
        self.exceptions = []
        self.dependencies = set()
        self.dependencies_headers = set()
        self.namespaces = set()

        for h in self.headers:
            self.classes.extend(h.classes.values())
            self.class_templates.extend(h.class_templates.values())
            self.typedefs.extend(h.typedefs)
            self.enums.extend(h.enums)
            self.functions.extend(h.functions)
            self.operators.extend(h.operators)
            self.class_dict.update(h.class_dict)
            self.class_template_dict.update(h.class_template_dict)
            self.typedef_dict.update(h.typedef_dict)
            self.dependencies_headers.update(h.dependencies)
            self.namespaces.update(h.namespaces)

        # clean up dependencies
        dependencies_clean = set()
        for d in self.dependencies_headers:
            name = self.get_module_name(d)
            if name in module_names:
                dependencies_clean.add(name)

        self.dependencies_headers = dependencies_clean - {self.name}

        self.sort_classes()

    def sort_classes(self):

        class_dict = {c.name: c for c in self.classes}
        dag = {
            c.name: set(s for s in c.superclass if s in class_dict)
            for c in self.classes
        }

        self.classes = [class_dict[el] for el in toposort_flatten(dag)]


if __name__ == "__main__":

    from os import getenv
    from .utils import init_clang

    init_clang()

    conda_prefix = Path(getenv("CONDA_PREFIX"))
    p = Path(conda_prefix / "include" / "opencascade")

    gp = ModuleInfo("gp", p, p.files("gp_*.hxx"))
    for el in [Path(el).split()[-1] for el in gp.dependencies]:
        print(el)

    TColStd = ModuleInfo("TColStd", p, p.files("TColStd_*.hxx"))
    for el in [Path(el).split()[-1] for el in gp.dependencies]:
        print(el)

    Standard = ModuleInfo("Standard", p, p.files("Standard_*.hxx"))
    for el in [Path(el).split()[-1] for el in gp.dependencies]:
        print(el)



from setuptools import setup

setup(
    name="pywrap",
    version="0.1dev",
    entry_points={"console_scripts": "pywrap = bindgen.__main__:main"},
    packages=["bindgen"],
    include_package_data=True,
    install_requires=[
        "click",
        "logzero",
        "path",
        "clang",
        "toml",
        "pandas",
        "joblib",
        "tqdm",
        "jinja2",
        "toposort",
        "pyparsing",
        "pybind11",
        "schema",
    ],
)



from functools import reduce
from operator import add
from re import match
from sys import platform
from typing import List

import logzero
import toml as toml
import pandas as pd
import pyparsing as pp

from joblib import Parallel, delayed
from path import Path
from tqdm import tqdm
from jinja2 import Environment, FileSystemLoader
from toposort import toposort_flatten

from .module import ModuleInfo
from .header import parse_tu, ClassInfo, get_symbols, get_namespaces
from .utils import current_platform, get_includes, init_clang
from .schemas import global_schema, module_schema


def read_settings(p):

    with open(p) as f:
        settings = toml.load(f)

    # validate
    settings = global_schema.validate(settings)

    # extract and compile the module name extraction callable
    code = compile("func={}".format(
        settings.pop("module_mapping")), "<tmp>", "exec")
    tmp = {"Path": Path}
    exec(code, tmp, tmp)
    module_mapping = tmp["func"]

    # get module specific settings if defined
    if "Modules" in settings:
        modules = settings.pop("Modules")
    else:
        modules = None

    return settings, module_mapping, modules


def read_symbols(p):
    """Read provided symbols file and return a dataframe

    This information is used later for flagging undefined symbols
    """

    if int(pd.__version__.split(".")[0]) >= 2:
        sym = pd.read_csv(
            p, header=None, names=["name"], sep="\\s+", on_bad_lines="skip"
        ).dropna()
    else:
        sym = pd.read_csv(
            p, header=None, names=["name"], delim_whitespace=True, error_bad_lines=False
        ).dropna()
    return sym


def remove_undefined_mangled(m, sym):

    # exclude methods
    for c in m.classes:
        c.methods_unfiltered = c.methods
        c.methods_byref_unfiltered = c.methods_byref
        c.methods_return_byref_unfiltered = c.methods_return_byref
        c.static_methods_unfiltered = c.static_methods
        c.static_methods_byref_unfiltered = c.static_methods_byref
        c.constructors_unfiltered = c.constructors

        c.methods = [
            el
            for el in c.methods
            if sym.name.str.endswith(el.mangled_name).any()
            or el.inline
            or el.pure_virtual
            or el.virtual
        ]
        c.methods_byref = [
            el
            for el in c.methods_byref
            if sym.name.str.endswith(el.mangled_name).any()
            or el.inline
            or el.pure_virtual
            or el.virtual
        ]
        c.methods_return_byref = [
            el
            for el in c.methods_return_byref
            if sym.name.str.endswith(el.mangled_name).any()
            or el.inline
            or el.pure_virtual
            or el.virtual
        ]
        c.static_methods = [
            el
            for el in c.static_methods
            if sym.name.str.endswith(el.mangled_name).any() or el.inline
        ]
        c.static_methods_byref = [
            el
            for el in c.static_methods_byref
            if sym.name.str.endswith(el.mangled_name).any() or el.inline
        ]
        c.constructors = [
            el
            for el in c.constructors
            if sym.name.str.endswith(el.mangled_name).any()
            or el.inline
            or el.pure_virtual
            or el.virtual
        ]

    # exclude functions
    m.functions = [
        f
        for f in m.functions
        if sym.name.str.startswith(f.mangled_name).any() or f.inline
    ]

    # exclude functions per header
    for h in m.headers:
        h.functions_unfiltered = h.functions
        h.functions = [
            f
            for f in h.functions
            if sym.name.str.startswith(f.mangled_name).any() or f.inline
        ]


def is_byref_arg(arg, byref_types):

    rv = False

    if any(arg.startswith(byref_t) and arg.endswith("&") for byref_t in byref_types):
        rv = True

    return rv


def is_byref_return(met):

    rv = False

    ret = met.return_type
    if ret.endswith("&") and len(met.args) == 0:
        rv = True

    return rv


def is_byref(met, byref_types):

    rv = False

    if met.return_type == "void":
        for arg in met.args:
            if any(
                arg.arg_type.startswith(byref_t) and arg.arg_type.endswith("&") for byref_t in byref_types
            ):
                rv = True
                break

    return rv


def type_form_byref_smart_ptr(t: str, ptr_types: List[str]) -> str:

    expr = (pp.Suppress(pp.Or(map(pp.Literal, ptr_types)) + pp.Literal('<'))
            + pp.Word(pp.alphanums+'_') + pp.Suppress(pp.Literal('>') + pp.Literal('&')))

    return expr.parse_string(t)[0]


def _exclude_methods(classes, exclusions):

    for pat in exclusions:
        cls_pat, m_pat = pat.split("::")
        for c in (c for c in classes if match(cls_pat, c.name)):
            c.methods = [m for m in c.methods if not match(m_pat, m.name)]
            c.static_methods = [
                m for m in c.static_methods if not match(m_pat, m.name)]
            c.operators = [m for m in c.operators if not match(m_pat, m.name)]


def transform_module(m, sym, settings, settings_per_module, platform=None):

    s = settings_per_module.get(m.name, None)
    global_excludes = settings[platform if platform else current_platform()]["exclude_classes"]

    # handle global excludes
    m.classes = [
        c for c in m.classes if not any(match(pat, c.name) for pat in global_excludes)
    ]
    m.class_dict = {
        k: v
        for k, v, in m.class_dict.items()
        if not any(match(pat, k) for pat in global_excludes)
    }

    if s:
        # exclude classes
        m.classes = [
            c for c in m.classes if c.name not in s["exclude_classes"]]
        m.class_dict = {
            k: v for k, v, in m.class_dict.items() if k not in s["exclude_classes"]
        }

        # exclude class templates
        m.class_templates = [
            c for c in m.class_templates if c.name not in s["exclude_class_templates"]
        ]
        m.class_template_dict = {
            k: v
            for k, v, in m.class_template_dict.items()
            if not any(match(f"^{pat}<.*>", k) for pat in s["exclude_class_templates"])
        }

        # exclude methods (including static methods)
        _exclude_methods(m.classes, s["exclude_methods"])
        _exclude_methods(m.class_templates,
                         s["exclude_class_template_methods"])

        # exclude functions
        m.functions = [
            f for f in m.functions if f.name not in s["exclude_functions"]]
        for h in m.headers:
            h.functions = [
                f for f in h.functions if f.name not in s["exclude_functions"]
            ]

        # exclude typedefs
        m.typedefs = [
            t for t in m.typedefs if t.name not in s["exclude_typedefs"]]
        for h in m.headers:
            h.typedefs = [
                t for t in h.typedefs if t.name not in s["exclude_typedefs"]]

    # collect methods and static methods using byref i.s.o. return
    byref_types = settings["byref_types"] + settings["byref_types_smart_ptr"]

    if byref_types:
        for c in m.classes:

            c.methods_byref = [
                met for met in c.methods if is_byref(met, byref_types)]
            c.methods_return_byref = [
                met
                for met in c.methods
                if is_byref_return(met) and not met.pure_virtual
            ]
            c.static_methods_byref = [
                met for met in c.static_methods if is_byref(met, byref_types)
            ]

            for met in c.methods_byref:
                c.methods.remove(met)

            for met in c.methods_return_byref:
                c.methods.remove(met)

            for met in c.static_methods_byref:
                c.static_methods.remove(met)

    # collect exceptions
    for c in m.classes:
        if any([match(pat, c.name) for pat in settings["exceptions"]]):
            m.exceptions.append(c)
        elif c.superclasses:
            for s in (s for s in c.superclasses if s):  # remove None etc
                if any([match(pat, s) for pat in settings["exceptions"]]):
                    m.exceptions.append(c)
                    break

    for ex in m.exceptions:
        m.classes.remove(ex)

    # remove undefined symbols
    remove_undefined_mangled(m, sym)

    # sort to be reproducible
    m.headers = sorted(m.headers, key=lambda x: x.name)
    for h in m.headers:
        h.dependencies = sorted(set(h.dependencies))


def split_into_modules(names, files):

    rv = {}

    for n in names:
        rv[n] = [
            f for f in files if f.name.startswith(n + "_") or f.name.startswith(n + ".")
        ]

    return rv


def parse_modules(
    verbose, n_jobs, settings, module_mapping, settings_per_module, target_platform
):

    if target_platform is None:
        target_platform = current_platform()

    settings["Modules"] = settings_per_module

    path = Path(settings["input_folder"])
    file_pats = settings["pats"]
    file_exc = settings["exclude"]
    module_names = settings["modules"]
    exclude_ns = settings["exclude_namespaces"]

    module_names += settings[target_platform]["modules"]

    file_pats = [p.format(m) for m in module_names for p in settings["pats"]]

    all_files = reduce(add, (path.files(pat) for pat in file_pats))
    all_files = set(f for f in all_files if f.name not in file_exc)

    module_dict = split_into_modules(module_names, all_files)

    modules = []

    # parse modules using libclang

    def _process_module(name, files, module_names, includes, clang_location):

        # loky based workaround
        get_includes.__defaults__ = includes
        init_clang.__defaults__ = clang_location

        # inject excluded namespaces
        get_symbols.__defaults__ = get_symbols.__defaults__[
            :-1] + (exclude_ns,)
        get_namespaces.__defaults__ = get_namespaces.__defaults__[
            :-1] + (exclude_ns,)

        if not verbose:
            logzero.logger.setLevel(logzero.logging.INFO)
        return ModuleInfo(name, path, files, module_names, settings, target_platform)

    modules = Parallel(prefer="processes", n_jobs=n_jobs)(
        delayed(_process_module)(
            name,
            files,
            module_names,
            get_includes.__defaults__,
            init_clang.__defaults__,
        )
        for name, files in tqdm(module_dict.items())
    )

    return modules


def transform_modules(
    verbose, n_jobs, settings, module_mapping, settings_per_module, modules, platform=None
):

    sym = read_symbols(settings[platform if platform else current_platform()]["symbols"])

    # ignore functions and classes based on settings and update the global class_dict
    def _filter_module(m):
        if not verbose:
            logzero.logger.setLevel(logzero.logging.INFO)
        logzero.logger.debug(m.name)
        transform_module(m, sym, settings, settings_per_module)

        return m

    modules = Parallel(prefer="processes", n_jobs=n_jobs)(
        delayed(_filter_module)(m) for m in tqdm(modules)
    )

    # construct global class dictionary
    class_dict = {}
    for m in modules:
        class_dict.update(m.class_dict)

    # sort modules
    logzero.logger.info("sorting")

    cls_dict = {c.name: m.name for m in modules for c in m.classes}
    enum_dict = {e.name: m.name for m in modules for e in m.enums}
    # Update dependencies based on superclasses and default argument types

    for m in modules:
        m.dependencies.update(
            [
                cls_dict[s]
                for c in m.classes
                for s in c.superclass
                if s in cls_dict and cls_dict[s] != m.name
            ]
        )
        # Consts should be removed before
        for t in (
            t
            for c in m.classes
            for method in c.methods + c.constructors + c.destructors
            for t in method.default_value_types
        ):
            if t.startswith("const "):
                t = t.split("const ")[1]
            if t in cls_dict and cls_dict[t] != m.name:
                m.dependencies.add(cls_dict[t])
            # elif t in enum_dict and enum_dict[t] != m.name:
            #    m.dependencies.add(enum_dict[t])

    # remove duplicate typedefs
    logzero.logger.info("removing duplicate typedefs")

    typedefs_dict = {}
    for m in modules:
        for h in m.headers:
            to_remove = []
            for t in h.typedefs:
                if t.type in typedefs_dict:
                    typedefs_dict[t.type].append(t)
                    to_remove.append(t)
                else:
                    typedefs_dict[t.type] = [t]

            for t in to_remove:
                h.typedefs.remove(t)

    return modules, class_dict, enum_dict


def toposort_modules(modules):

    deps = {}

    cls_dict = {c.name: m.name for m in modules for c in m.classes}
    tmpl_dict = {t.name: t for m in modules for t in m.class_templates}

    for m in modules:

        typedefs = []
        for t in m.typedefs:
            if not t.pod:
                if t.template_base:
                    if t.template_base[0] in tmpl_dict:
                        typedefs.append(tmpl_dict[t.template_base[0]])

        deps[m.name] = set(
            cls_dict[s]
            for c in m.classes + m.class_templates + typedefs
            for s in c.superclass
            if s in cls_dict
        ) - {m.name}

    return toposort_flatten(deps)


def render(settings, module_settings, modules, class_dict, prefix=Path(""), platform=None):

    name = settings["name"]
    module_names = [m.name for m in modules]
    output_path = Path(settings["output_folder"])
    operator_dict = settings["Operators"]

    pre = settings["Extras"]["include_pre"]
    post = settings["Extras"]["include_post"]

    def proper_new_operator(cls):

        new_ops = [op for op in cls.static_operators if op.name == "operator new"]

        if not new_ops:
            return True

        new_ops = [op for op in new_ops if len(op.args) == 1]

        return new_ops

    def proper_delete_operator(cls):

        del_ops = [op for op in cls.static_operators if op.name ==
                   "operator delete"]

        if not del_ops:
            return True

        del_ops = [op for op in del_ops if len(op.args) == 1]

        return del_ops

    default_path = [Path(__file__).dirname()]
    additional_path = (
        [Path(settings["template_path"]),] if settings["template_path"] else []
    )
    template_paths = additional_path + default_path

    jinja_env = Environment(
        loader=FileSystemLoader(template_paths),
        trim_blocks=True,
        lstrip_blocks=True,
        extensions=["jinja2.ext.do"],
    )

    all_classes = {c.name: c for m in modules for c in m.classes}
    all_enums = {e.name: e for m in modules for e in m.enums}
    all_typedefs = {t.name: t for m in modules for t in m.typedefs}

    jinja_env.globals.update(
        {
            "parent_has_nonpublic_destructor": lambda c: any(
                all_classes[p].nonpublic_destructors
                for p in c.superclasses
                if p in all_classes
            ),
            "is_byref": lambda t: is_byref_arg(t, settings["byref_types"]),
            "is_byref_smart_ptr": lambda t: is_byref_arg(
                t, settings["byref_types_smart_ptr"]
            ),
            "type_from_byref_smart_ptr": lambda t: type_form_byref_smart_ptr(t, settings["byref_types_smart_ptr"]),
            "args_byref": lambda f: [
                arg for arg, t, _, _  in f.args if is_byref_arg(t, settings["byref_types"])
            ],
            "enumerate": enumerate,
            "platform": platform if platform else current_platform(),
            "class_dict": class_dict,
            "all_classes": all_classes,
            "all_enums": all_enums,
            "all_typedefs": all_typedefs,
            "project_name": name,
            "operator_dict": operator_dict,
            "include_pre": pre,
            "include_post": post,
            "references_inner": lambda name, method: name + "::" in method.return_type
            or any([name + "::" in arg.arg_type for arg in method.args]),
            "proper_new_operator": proper_new_operator,
            "proper_delete_operator": proper_delete_operator,
            "module_names": module_names,
            "sorted_modules": toposort_modules(modules),
            "settings": settings,
        }
    )

    jinja_env.filters["get"] = lambda x, n: x[n]

    template_sub = jinja_env.get_template("template_sub.j2")
    template_sub_pre = jinja_env.get_template("template_sub_pre.j2")
    template_tmpl = jinja_env.get_template("template_templates.j2")
    template_main = jinja_env.get_template("template_main.j2")
    template_cmake = jinja_env.get_template("CMakeLists.j2")

    output_path.mkdir_p()
    with output_path:
        for m in tqdm(modules):
            tqdm.write(f"Processing module {m.name}")

            jinja_env.globals.update(
                {
                    "module_settings": module_settings.get(
                        m.name, module_schema.validate({})
                    )
                }
            )

            class_templates = {el.name: el for el in m.class_templates}

            typedefs = (
                t
                for h in m.headers
                for t in h.typedefs
                if not t.pod
                and not "_H" in t.name
                and len(t.template_base) > 0
                and not t.type.startswith("opencascade::handle")
            )
            typedefs = filter(
                lambda t: not t.type.split(t.template_base[0])[-1].endswith(
                    "::Iterator"
                ),
                typedefs,
            )

            classes_typedefs = {el.name: el for el in (
                m.classes + list(typedefs))}

            dag = {}
            for el in classes_typedefs.values():
                if isinstance(el, ClassInfo):
                    deps = set(el.superclass)
                else:
                    base = el.template_base[0]
                    deps = set(el.template_args)
                    deps |= (
                        set(class_templates[base].superclass)
                        if base in class_templates
                        else set()
                    )

                dag[el.name] = deps

            sorted_classes_typedefs = [
                classes_typedefs[k]
                for k in toposort_flatten(dag)
                if k in classes_typedefs
            ]

            with open(f"{m.name}_pre.cpp", "w") as f:
                f.write(
                    template_sub_pre.render(
                        {
                            "module": m,
                            "classes_typedefs": sorted_classes_typedefs,
                            "str": str,
                            "type": type,
                        }
                    )
                )

            with open(f"{m.name}.cpp", "w") as f:
                f.write(template_sub.render({"module": m}))

            with open(f"{m.name}_tmpl.hxx", "w") as f:
                f.write(template_tmpl.render({"module": m}))

        with open(f"{name}.cpp", "w") as f:
            f.write(template_main.render({"name": name}))

        with open("CMakeLists.txt", "w") as f:
            f.write(template_cmake.render({"name": name}))

    for p in settings["additional_files"]:
        (prefix / Path(p)).copy(output_path)


def validate_result(verbose, n_jobs, folder):
    def _validate(f):

        tu = parse_tu(f, pre_includes="")
        if len([d for d in tu.diagnostics if d.severity > 2]):
            logzero.logger.error("Validation {}: NOK".format(f))
            if verbose:
                for d in tu.diagnostics:
                    logzero.logger.warning(d)
        else:
            logzero.logger.info("Validation {}: OK".format(f))

    result = Parallel(prefer="processes", n_jobs=n_jobs)(
        delayed(_validate)(n) for n in Path(folder).files("*.cpp")
    )

    for r in result:
        pass
